#include "imgui_manager.hpp"

ImGuiManager::~ImGuiManager() { cleanup(); }

void ImGuiManager::init(
  vk::Instance            instance,
  vk::PhysicalDevice      physicalDevice,
  vk::Device              device,
  uint32_t                queueFamily,
  vk::Queue               queue,
  vk::RenderPass          renderPass,
  uint32_t                minImageCount,
  uint32_t                imageCount,
  vk::SampleCountFlagBits msaaSamples,
  GLFWwindow*             window
) {
  mDevice = device;

  // Create ImGui context
  IMGUI_CHECKVERSION();
  ImGui::CreateContext();

  ImGuiIO& imGuiIO = ImGui::GetIO();
  imGuiIO.ConfigFlags |= ImGuiConfigFlags_DockingEnable;
  imGuiIO.ConfigFlags |= ImGuiConfigFlags_ViewportsEnable;
  imGuiIO.IniFilename = nullptr; // Disable writing imgui.ini

  // Setup Dear ImGui style
  ImGui::StyleColorsDark();

  // Initialize ImGui for GLFW and Vulkan
  ImGui_ImplGlfw_InitForVulkan(window, true);

  vk::DescriptorPoolSize descriptorPoolSize {
    .type            = vk::DescriptorType::eCombinedImageSampler,
    .descriptorCount = 1,
  };

  vk::DescriptorPoolCreateInfo poolInfo {
    .flags         = vk::DescriptorPoolCreateFlagBits::eFreeDescriptorSet,
    .maxSets       = 1,
    .poolSizeCount = 1,
    .pPoolSizes    = &descriptorPoolSize,
  };

  mDescriptorPool = device.createDescriptorPoolUnique(poolInfo);

  ImGui_ImplVulkan_InitInfo initInfo {
    .Instance                    = instance,
    .PhysicalDevice              = physicalDevice,
    .Device                      = device,
    .QueueFamily                 = queueFamily,
    .Queue                       = queue,
    .DescriptorPool              = mDescriptorPool.get(),
    .RenderPass                  = renderPass,
    .MinImageCount               = minImageCount,
    .ImageCount                  = imageCount,
    .MSAASamples                 = static_cast<VkSampleCountFlagBits>(msaaSamples),
    .PipelineCache               = VK_NULL_HANDLE,
    .Subpass                     = 0,
    .UseDynamicRendering         = false,
    .PipelineRenderingCreateInfo = {},
    .Allocator                   = nullptr,
    .CheckVkResultFn             = nullptr,
    .MinAllocationSize           = 0,
  };

  ImGui_ImplVulkan_Init(&initInfo);
  mInitialized = true;
}

void ImGuiManager::cleanup() {
  if (mInitialized) {
    ImGui_ImplVulkan_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();
    mInitialized = false;
  }
}

void ImGuiManager::newFrame() {
  ImGui_ImplVulkan_NewFrame();
  ImGui_ImplGlfw_NewFrame();
  ImGui::NewFrame();
  ImGui::Begin("Controls");
}

void ImGuiManager::render() {
  ImGui::End();
  ImGui::Render();

  if (ImGui::GetIO().ConfigFlags & ImGuiConfigFlags_ViewportsEnable) {
    ImGui::UpdatePlatformWindows();
    ImGui::RenderPlatformWindowsDefault();
  }
}

void ImGuiManager::renderControls(
  float&     cameraSpeed,
  float&     fieldOfView,
  bool&      wireframeMode,
  float&     lineWidth,
  float      maxLineWidth,
  bool       wideLineSupport,
  Camera&    camera,
  LightInfo& lightSettings,
  bool&      needsPipelineRecreation
) {
  // Camera Settings
  if (ImGui::CollapsingHeader("Camera Settings", ImGuiTreeNodeFlags_DefaultOpen)) {
    ImGui::SliderFloat("Camera Speed", &cameraSpeed, 0.1F, 10.0F, "%.1f");
    ImGui::SliderFloat("Field of View", &fieldOfView, 45.0F, 120.0F, "%.1f");
    if (ImGui::Button("Reset Camera"))
      camera = Camera();
  }

  // Rendering Settings
  if (ImGui::CollapsingHeader("Rendering Settings", ImGuiTreeNodeFlags_DefaultOpen)) {
    if (ImGui::Checkbox("Wireframe Mode", &wireframeMode))
      needsPipelineRecreation = true;

    if (wireframeMode) {
      if (wideLineSupport) {
        if (ImGui::SliderFloat("Line Width", &lineWidth, 1.0F, maxLineWidth, "%.1f")) {
          needsPipelineRecreation = true;
        }
      } else {
        ImGui::TextColored(ImVec4(1.0F, 0.5F, 0.5F, 1.0F), "Wide lines not supported on this device");
        lineWidth = 1.0F;
      }
    }
  }

  renderControlsHelp();
  renderFrameMetrics();
  renderCameraInfo(camera);
  renderLightControls(lightSettings);
}

void ImGuiManager::renderControlsHelp() {
  if (ImGui::CollapsingHeader("Controls", ImGuiTreeNodeFlags_DefaultOpen)) {
    ImGui::BulletText("Use mouse to look around");
    ImGui::BulletText("WASD to move horizontally");
    ImGui::BulletText("Space/Shift to move up/down");
    ImGui::BulletText("ESC to toggle mouse capture");
  }
}

void ImGuiManager::renderFrameMetrics() {
  if (ImGui::CollapsingHeader("Performance", ImGuiTreeNodeFlags_DefaultOpen)) {
    ImGui::Text(
      "Application average %.3f ms/frame (%.1f FPS)",
      static_cast<double>(1000.0F / ImGui::GetIO().Framerate),
      static_cast<double>(ImGui::GetIO().Framerate)
    );
  }
}

void ImGuiManager::renderMemoryUsage(
  size_t vertexSize,
  size_t indexSize,
  size_t vertexCount,
  size_t indexCount
) {
  ImGui::Separator();
  ImGui::Text("Memory Usage:");
  ImGui::Text(
    "Vertex Buffer: %.2f MB",
    (static_cast<double>(vertexCount) * static_cast<double>(vertexSize)) / (1024.0 * 1024.0)
  );
  ImGui::Text(
    "Index Buffer: %.2f MB",
    (static_cast<double>(indexCount) * static_cast<double>(indexSize)) / (1024.0 * 1024.0)
  );
}

void ImGuiManager::renderCameraInfo(const Camera& camera) {
  ImGui::Separator();
  ImGui::Text("Camera Information:");
  ImGui::Text(
    "Position: (%.2f, %.2f, %.2f)", camera.getPosition().x, camera.getPosition().y, camera.getPosition().z
  );
  ImGui::Text(
    "Front Vector: (%.2f, %.2f, %.2f)", camera.getFront().x, camera.getFront().y, camera.getFront().z
  );
  ImGui::Text("Yaw: %.2f°, Pitch: %.2f°", camera.getYaw(), camera.getPitch());
}

void ImGuiManager::renderLightControls(LightInfo& lightSettings) {
  ImGui::Separator();
  ImGui::Text("Light Controls:");

  // Light Position
  ImGui::DragFloat3("Light Position", &lightSettings.position.x, 0.1F, -10.0F, 10.0F);

  // Light Color
  ImGui::ColorEdit3("Light Color", &lightSettings.color.x);

  // Light Strengths
  ImGui::SliderFloat("Ambient Strength", &lightSettings.ambient_strength, 0.0F, 1.0F);
  ImGui::SliderFloat("Specular Strength", &lightSettings.specular_strength, 0.0F, 1.0F);
}
