#include "debug_messenger.hpp"

namespace {
  constexpr std::string_view ANSI_RESET      = "\033[0m";
  constexpr std::string_view ANSI_BOLD       = "\033[1m";
  constexpr std::string_view ANSI_RED        = "\033[31m";
  constexpr std::string_view ANSI_YELLOW     = "\033[33m";
  constexpr std::string_view ANSI_LIGHT_BLUE = "\033[94m";
  constexpr std::string_view ANSI_WHITE      = "\033[97m";
}

fn DebugMessenger::create(const vk::Instance& instance, Config config)
  -> std::expected<vk::UniqueDebugUtilsMessengerEXT, std::string> {
#ifdef NDEBUG
  return nullptr;
#else
  try {
    vk::DebugUtilsMessengerCreateInfoEXT createInfo {
      .messageSeverity = config.severity_flags,
      .messageType     = config.type_flags,
      .pfnUserCallback = debugCallback,
      .pUserData       = &config,
    };

    return instance.createDebugUtilsMessengerEXTUnique(createInfo);
  } catch (const vk::SystemError& e) {
    return std::unexpected(std::format("Failed to create debug messenger: {}", e.what()));
  }
#endif
}

VKAPI_ATTR VkBool32 VKAPI_CALL DebugMessenger::debugCallback(
  VkDebugUtilsMessageSeverityFlagBitsEXT      messageSeverity,
  VkDebugUtilsMessageTypeFlagsEXT             messageType,
  const VkDebugUtilsMessengerCallbackDataEXT* pCallbackData,
  void*                                       pUserData
) {
  const auto* config = static_cast<const Config*>(pUserData);

  Message msg { .message       = pCallbackData->pMessage,
                .severity      = static_cast<vk::DebugUtilsMessageSeverityFlagBitsEXT>(messageSeverity),
                .type          = static_cast<vk::DebugUtilsMessageTypeFlagsEXT>(messageType),
                .function_name = pCallbackData->pMessageIdName
                                   ? std::optional<std::string_view>(pCallbackData->pMessageIdName)
                                   : std::nullopt,
                .id            = pCallbackData->messageIdNumber };

  const auto formattedMessage = formatMessage(msg);

  if (config && config->use_stderr_for_errors &&
      (messageSeverity & VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT)) {
    std::println(stderr, "{}", formattedMessage);
  } else {
    std::println("{}", formattedMessage);
  }

  return VK_FALSE;
}

fn DebugMessenger::formatMessage(const Message& msg) -> std::string {
  // Color based on severity
  std::string_view colorCode = ANSI_WHITE;
  std::string_view severityText;

  switch (msg.severity) {
    case vk::DebugUtilsMessageSeverityFlagBitsEXT::eVerbose:
      colorCode    = ANSI_LIGHT_BLUE;
      severityText = "VERBOSE";
      break;
    case vk::DebugUtilsMessageSeverityFlagBitsEXT::eInfo:
      colorCode    = ANSI_WHITE;
      severityText = "INFO";
      break;
    case vk::DebugUtilsMessageSeverityFlagBitsEXT::eWarning:
      colorCode    = ANSI_YELLOW;
      severityText = "WARNING";
      break;
    case vk::DebugUtilsMessageSeverityFlagBitsEXT::eError:
      colorCode    = ANSI_RED;
      severityText = "ERROR";
      break;
  }

  // Build message type string
  std::string typeStr;
  if (msg.type & vk::DebugUtilsMessageTypeFlagBitsEXT::eGeneral)
    typeStr += "GENERAL ";
  if (msg.type & vk::DebugUtilsMessageTypeFlagBitsEXT::eValidation)
    typeStr += "VALIDATION ";
  if (msg.type & vk::DebugUtilsMessageTypeFlagBitsEXT::ePerformance)
    typeStr += "PERFORMANCE ";

  // Format the message with color and structure
  return std::format(
    "{}{}[{}] {} {}{}: {}{}",
    colorCode,
    ANSI_BOLD,
    severityText,
    typeStr,
    msg.function_name.has_value() ? std::format("in {} ", *msg.function_name) : "",
    msg.id,
    msg.message,
    ANSI_RESET
  );
}
