#pragma once

#define VULKAN_HPP_DISPATCH_LOADER_DYNAMIC 1 // Use dynamic dispatch for Vulkan functions
#define VULKAN_HPP_NO_CONSTRUCTORS
#include <expected>
#include <string>
#include <string_view>
#include <vulkan/vulkan.hpp>

#include "../util/types.hpp"

class DebugMessenger {
 public:
  /**
   * @brief Configuration for debug messenger
   */
  struct Config {
    vk::DebugUtilsMessageSeverityFlagsEXT severity_flags;
    vk::DebugUtilsMessageTypeFlagsEXT     type_flags;
    bool                                  use_stderr_for_errors;
  };

  /**
   * @brief Debug message information
   */
  struct Message {
    std::string_view                         message;
    vk::DebugUtilsMessageSeverityFlagBitsEXT severity;
    vk::DebugUtilsMessageTypeFlagsEXT        type;
    std::optional<std::string_view>          function_name;
    i32                                      id;
  };

  DebugMessenger()                                     = default;
  DebugMessenger(const DebugMessenger&)                = default;
  DebugMessenger(DebugMessenger&&)                     = delete;
  fn operator=(const DebugMessenger&)->DebugMessenger& = default;
  fn operator=(DebugMessenger&&)->DebugMessenger&      = delete;
  ~DebugMessenger()                                    = default;

  /**
   * @brief Creates a debug messenger with the specified configuration
   * @param instance Vulkan instance
   * @param config Debug messenger configuration
   * @return Expected containing the debug messenger or an error message
   */
  static fn create(
    const vk::Instance& instance,
    Config              config = { .severity_flags = vk::DebugUtilsMessageSeverityFlagBitsEXT::eVerbose |
                                                     vk::DebugUtilsMessageSeverityFlagBitsEXT::eWarning |
                                                     vk::DebugUtilsMessageSeverityFlagBitsEXT::eError,
                                   .type_flags = vk::DebugUtilsMessageTypeFlagBitsEXT::eGeneral |
                                                 vk::DebugUtilsMessageTypeFlagBitsEXT::eValidation |
                                                 vk::DebugUtilsMessageTypeFlagBitsEXT::ePerformance,
                                   .use_stderr_for_errors = true }
  ) -> std::expected<vk::UniqueDebugUtilsMessengerEXT, std::string>;

 private:
  static VKAPI_ATTR VkBool32 VKAPI_CALL debugCallback(
    VkDebugUtilsMessageSeverityFlagBitsEXT      messageSeverity,
    VkDebugUtilsMessageTypeFlagsEXT             messageType,
    const VkDebugUtilsMessengerCallbackDataEXT* pCallbackData,
    void*                                       pUserData
  );

  /**
   * @brief Formats a debug message for output
   * @param msg Message information
   * @return Formatted message string
   */
  static fn formatMessage(const Message& msg) -> std::string;
};
