#pragma once

#include <expected>
#include <span>

#define VULKAN_HPP_NO_CONSTRUCTORS
#include <vulkan/vulkan.hpp>

#include "../structs/queue_family_indices.hpp"
#include "../util/types.hpp"

class DeviceManager {
 public:
  DeviceManager() = default;

  /**
   * @brief Result type for physical device selection
   */
  struct PhysicalDeviceInfo {
    vk::PhysicalDevice      physical_device;
    vk::SampleCountFlagBits msaa_samples;
    f32                     max_line_width;
    bool                    wide_line_support;
  };

  /**
   * @brief Result type for logical device creation
   */
  struct LogicalDeviceInfo {
    vk::UniqueDevice device;
    vk::Queue        graphics_queue;
    vk::Queue        present_queue;
  };

  /**
   * @brief Picks a suitable physical device (GPU) for the application.
   *
   * @param instance Vulkan instance to use for device selection
   * @param surface Surface to check for compatibility
   * @param requiredExtensions List of required device extensions
   * @return std::expected<PhysicalDeviceInfo, std::string> Physical device info or error message
   */
  static fn pickPhysicalDevice(
    const vk::Instance&          instance,
    const vk::SurfaceKHR&        surface,
    std::span<const char* const> requiredExtensions
  ) -> std::expected<PhysicalDeviceInfo, std::string>;

  /**
   * @brief Creates a logical device with the specified features.
   *
   * @param physicalDevice Physical device to create logical device from
   * @param surface Surface to create device with
   * @param requiredExtensions Device extensions to enable
   * @param enableValidationLayers Whether to enable validation layers
   * @param validationLayers Validation layers to enable if enabled
   * @return std::expected<LogicalDeviceInfo, std::string> Logical device info or error message
   */
  static fn createLogicalDevice(
    const vk::PhysicalDevice&    physicalDevice,
    const vk::SurfaceKHR&        surface,
    std::span<const char* const> requiredExtensions,
    bool                         enableValidationLayers,
    std::span<const char* const> validationLayers
  ) -> std::expected<LogicalDeviceInfo, std::string>;

  /**
   * @brief Gets the maximum supported MSAA sample count.
   *
   * @param physicalDevice Physical device to check
   * @return vk::SampleCountFlagBits Maximum supported sample count
   */
  static fn getMaxUsableSampleCount(const vk::PhysicalDevice& physicalDevice) -> vk::SampleCountFlagBits;

  /**
   * @brief Finds queue families that support required operations.
   *
   * @param device Physical device to check
   * @param surface Surface to check presentation support
   * @return QueueFamilyIndices Indices of suitable queue families
   */
  static fn findQueueFamilies(const vk::PhysicalDevice& device, const vk::SurfaceKHR& surface)
    -> QueueFamilyIndices;

  /**
   * @brief Checks if a physical device is suitable for the application.
   *
   * @param device Physical device to check
   * @param surface Surface to check compatibility with
   * @param requiredExtensions Required device extensions
   * @return bool True if device is suitable
   */
  static fn isDeviceSuitable(
    const vk::PhysicalDevice&    device,
    const vk::SurfaceKHR&        surface,
    std::span<const char* const> requiredExtensions
  ) -> bool;

  /**
   * @brief Checks if the device supports the required extensions.
   */
  static fn checkDeviceExtensionSupport(
    const vk::PhysicalDevice&    device,
    std::span<const char* const> requiredExtensions
  ) -> bool;
};
