#pragma once

#define VKFW_NO_STRUCT_CONSTRUCTORS // Use aggregate initialization for GLFW structs
#include <vkfw.hpp>

#include "../util/types.hpp"

struct WindowCallbacks {
  std::function<void(const vkfw::Window&, f64, f64)> on_cursor_move;
  std::function<
    void(const vkfw::Window&, const vkfw::Key&, i32, const vkfw::KeyAction&, const vkfw::ModifierKeyFlags&)>
    on_key;
  std::function<
    void(const vkfw::Window&, const vkfw::MouseButton&, const vkfw::MouseButtonAction&, const vkfw::ModifierKeyFlags&)>
                                                         on_mouse_button;
  std::function<void(const vkfw::Window&, usize, usize)> on_window_resize;
};

class WindowManager {
 public:
  WindowManager() = default;

  /**
   * @brief Creates a window with the specified dimensions and title.
   *
   * @param title Window title
   * @param callbacks Window callbacks
   * @return vkfw::UniqueWindow Unique pointer to the created window
   */
  static fn create(const char* title, const WindowCallbacks& callbacks)
    -> std::tuple<vkfw::UniqueInstance, vkfw::UniqueWindow>;

  /**
   * @brief Gets the required instance extensions for window surface creation.
   *
   * @return std::vector<const char*> List of required extensions
   */
  static fn getRequiredExtensions() -> std::vector<const char*>;

  /**
   * @brief Gets the framebuffer size of the window.
   *
   * @param window GLFW window
   * @return std::pair<i32, i32> Width and height of the framebuffer
   */
  static fn getFramebufferSize(const vkfw::Window& window) -> std::pair<i32, i32>;

  /**
   * @brief Sets the window title.
   *
   * @param window GLFW window
   * @param title New window title
   */
  static fn setTitle(const vkfw::Window& window, const std::string& title) -> void;

  /**
   * @brief Sets the cursor mode.
   *
   * @param window GLFW window
   * @param mode Cursor mode (Normal or Disabled)
   */
  static fn setCursorMode(const vkfw::Window& window, vkfw::CursorMode mode) -> void;

  /**
   * @brief Sets the window position.
   *
   * @param window GLFW window
   * @param xpos X position
   * @param ypos Y position
   */
  static fn setPosition(const vkfw::Window& window, i32 xpos, i32 ypos) -> void;
};
